@@ -336,7 +336,7 @@ handle_msg_start_emulating(struct ei_device *device, uint32_t serial, uint32_t s
 	case EI_DEVICE_STATE_EMULATING:
 	case EI_DEVICE_STATE_REMOVED_FROM_SERVER:
 		result = brei_result_new(EI_CONNECTION_DISCONNECT_REASON_PROTOCOL,
-					 "Invalid device state %ud for a start_emulating event", device->state);
+					 "Invalid device state %u for a start_emulating event", device->state);
 		break;
 	case EI_DEVICE_STATE_RESUMED:
 		ei_queue_device_start_emulating_event(device, sequence);
@@ -366,7 +366,7 @@ handle_msg_stop_emulating(struct ei_device *device, uint32_t serial)
 	case EI_DEVICE_STATE_RESUMED:
 	case EI_DEVICE_STATE_REMOVED_FROM_SERVER:
 		result = brei_result_new(EI_CONNECTION_DISCONNECT_REASON_PROTOCOL,
-					 "Invalid device state %ud for a stop_emulating event", device->state);
+					 "Invalid device state %u for a stop_emulating event", device->state);
 		break;
 	case EI_DEVICE_STATE_EMULATING:
 		ei_queue_device_stop_emulating_event(device);
@@ -399,7 +399,7 @@ maybe_error_on_device_state(struct ei_device *device, const char *event_type)
 	}
 
 	return brei_result_new(EI_CONNECTION_DISCONNECT_REASON_PROTOCOL,
-			       "Invalid device state %ud for a %s event", device->state, event_type);
+			       "Invalid device state %u for a %s event", device->state, event_type);
 }
 
 static struct brei_result *
@@ -858,6 +858,31 @@ handle_msg_touch_up(struct ei_touchscreen *touchscreen, uint32_t touchid)
 	return maybe_error_on_device_state(device, "touch up");
 }
 
+static struct brei_result *
+handle_msg_touch_cancel(struct ei_touchscreen *touchscreen, uint32_t touchid)
+{
+	struct ei_device *device = ei_touchscreen_get_device(touchscreen);
+
+	DISCONNECT_IF_SENDER_CONTEXT(device);
+
+	if (!ei_device_has_capability(device, EI_DEVICE_CAP_TOUCH)) {
+		return brei_result_new(EI_CONNECTION_DISCONNECT_REASON_PROTOCOL,
+				       "Touch cancel event for non-touch device");
+	}
+
+	struct ei *ei = ei_device_get_context(device);
+	if (ei->interface_versions.ei_touchscreen < EI_TOUCHSCREEN_EVENT_CANCEL_SINCE_VERSION) {
+		return brei_result_new(EI_CONNECTION_DISCONNECT_REASON_PROTOCOL,
+				       "Touch cancel event for touchscreen version v1");
+	}
+
+	if (device->state == EI_DEVICE_STATE_EMULATING) {
+		ei_queue_touch_cancel_event(device, touchid);
+		return NULL;
+	}
+	return maybe_error_on_device_state(device, "touch cancel");
+}
+
 static struct brei_result *
 handle_msg_touchscreen_destroy(struct ei_touchscreen *touchscreen, uint32_t serial)
 {
@@ -875,6 +900,7 @@ static const struct ei_touchscreen_interface touchscreen_interface = {
 	.down = handle_msg_touch_down,
 	.motion = handle_msg_touch_motion,
 	.up = handle_msg_touch_up,
+	.cancel = handle_msg_touch_cancel,
 };
 
 const struct ei_touchscreen_interface *
@@ -1063,6 +1089,7 @@ ei_device_removed_by_server(struct ei_device *device)
 {
 	struct ei_seat *seat = ei_device_get_seat(device);
 	struct ei *ei = ei_device_get_context(device);
+	struct ei_event *event;
 
 	switch (device->state) {
 	case EI_DEVICE_STATE_NEW:
@@ -1073,6 +1100,11 @@ ei_device_removed_by_server(struct ei_device *device)
 	case EI_DEVICE_STATE_PAUSED:
 	case EI_DEVICE_STATE_RESUMED:
 	case EI_DEVICE_STATE_EMULATING:
+		list_for_each_safe(event, &device->pending_event_queue, link) {
+			list_remove(&event->link);
+			ei_event_unref(event);
+		}
+
 		/* in the case of ei_disconnect() we may fake the
 		 * removal by the server, so we need to also remove the
 		 * pointer/keyboard/touch interfaces
@@ -1642,6 +1674,21 @@ ei_send_touch_up(struct ei_device *device, uint32_t tid)
 	return rc;
 }
 
+static int
+ei_send_touch_cancel(struct ei_device *device, uint32_t tid)
+{
+	struct ei *ei = ei_device_get_context(device);
+
+	if (ei->state == EI_STATE_NEW || ei->state == EI_STATE_DISCONNECTED)
+		return 0;
+
+	device->send_frame_event = true;
+
+	int rc = ei_touchscreen_request_cancel(device->touchscreen, tid);
+	if (rc)
+		ei_disconnect(ei);
+	return rc;
+}
 
 _public_ struct ei_touch *
 ei_device_touch_new(struct ei_device *device)
@@ -1735,6 +1782,32 @@ ei_touch_up(struct ei_touch *touch)
 	ei_send_touch_up(touch->device, touch->tracking_id);
 }
 
+_public_ void
+ei_touch_cancel(struct ei_touch *touch)
+{
+	struct ei_device *device = ei_touch_get_device(touch);
+	if (device->state != EI_DEVICE_STATE_EMULATING) {
+		log_bug_client(ei_device_get_context(device),
+			       "%s: device is not emulating", __func__);
+		return;
+	}
+
+	if (touch->state != TOUCH_IS_DOWN) {
+		log_bug_client(ei_device_get_context(device),
+			       "%s: touch %u is not currently down", __func__, touch->tracking_id);
+		return;
+	}
+
+	touch->state = TOUCH_IS_UP;
+
+	struct ei *ei = ei_device_get_context(device);
+
+	if (ei->interface_versions.ei_touchscreen >= EI_TOUCHSCREEN_REQUEST_CANCEL_SINCE_VERSION)
+		ei_send_touch_cancel(touch->device, touch->tracking_id);
+	else
+		ei_send_touch_up(touch->device, touch->tracking_id);
+}
+
 _public_ void
 ei_device_frame(struct ei_device *device, uint64_t time)
 {
