@@ -486,6 +486,7 @@ new_context(enum peck_ei_mode ei_mode)
 	peck->eis = eis;
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_FRAME);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 
 	struct ei *ei = ei_mode == PECK_EI_RECEIVER ? ei_new_receiver(peck) : ei_new_sender(peck);
 	ei_set_user_data(ei, peck);
@@ -501,6 +502,7 @@ new_context(enum peck_ei_mode ei_mode)
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTOSEAT);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTOSTART);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_FRAME);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_SYNC);
 
 	peck_ei_enable_fatal_bug(peck);
 	peck_eis_enable_fatal_bug(peck);
@@ -598,6 +600,7 @@ peck_enable_eis_behavior(struct peck *peck, enum peck_eis_behavior behavior)
 	case PECK_EIS_BEHAVIOR_HANDLE_BIND_SEAT:
 	case PECK_EIS_BEHAVIOR_HANDLE_CLOSE_DEVICE:
 	case PECK_EIS_BEHAVIOR_HANDLE_FRAME:
+	case PECK_EIS_BEHAVIOR_HANDLE_SYNC:
 	case PECK_EIS_BEHAVIOR_HANDLE_START_EMULATING:
 	case PECK_EIS_BEHAVIOR_HANDLE_STOP_EMULATING:
 	case PECK_EIS_BEHAVIOR_ADD_POINTER:
@@ -620,7 +623,6 @@ peck_enable_eis_behavior(struct peck *peck, enum peck_eis_behavior behavior)
 		break;
 	}
 }
-
 void
 peck_enable_ei_behavior(struct peck *peck, enum peck_ei_behavior behavior)
 {
@@ -654,6 +656,7 @@ peck_enable_ei_behavior(struct peck *peck, enum peck_ei_behavior behavior)
 	case PECK_EI_BEHAVIOR_HANDLE_ADDED_BUTTON:
 	case PECK_EI_BEHAVIOR_HANDLE_ADDED_SCROLL:
 	case PECK_EI_BEHAVIOR_HANDLE_FRAME:
+	case PECK_EI_BEHAVIOR_HANDLE_SYNC:
 		flag_set(peck->ei_behavior, behavior);
 		break;
 	case PECK_EI_BEHAVIOR_HANDLE_RESUMED:
@@ -909,6 +912,12 @@ _peck_dispatch_eis(struct peck *peck, int lineno)
 			else
 				process_event = tristate_no;
 			break;
+		case EIS_EVENT_PONG:
+			break;
+		case EIS_EVENT_SYNC:
+			if (flag_is_set(peck->eis_behavior, PECK_EIS_BEHAVIOR_HANDLE_SYNC))
+				process_event = tristate_yes;
+			break;
 		case EIS_EVENT_FRAME:
 			/* Ensure we only expect frames when we expect them */
 			munit_assert_true(need_frame);
@@ -1067,6 +1076,12 @@ _peck_dispatch_ei(struct peck *peck, int lineno)
 			if (flag_is_set(peck->ei_behavior, PECK_EI_BEHAVIOR_HANDLE_PAUSED))
 				process_event = tristate_yes;
 			break;
+		case EI_EVENT_PONG:
+			break;
+		case EI_EVENT_SYNC:
+			if (flag_is_set(peck->ei_behavior, PECK_EI_BEHAVIOR_HANDLE_SYNC))
+				process_event = tristate_yes;
+			break;
 		case EI_EVENT_FRAME:
 			/* Ensure we only expect frames when we expect them */
 			munit_assert_true(need_frame);
@@ -1222,7 +1237,7 @@ peck_drain_ei(struct ei *ei)
 }
 
 void
-peck_assert_no_ei_events(struct ei *ei)
+_peck_assert_no_ei_events(struct ei *ei, int lineno)
 {
 	struct peck *peck = ei_get_user_data(ei);
 
@@ -1238,13 +1253,19 @@ peck_assert_no_ei_events(struct ei *ei)
 			continue;
 		}
 
-		munit_errorf("Expected empty event queue, have: %s\n",
-			     peck_ei_event_name(e));
+		if (peck && flag_is_set(peck->ei_behavior, PECK_EI_BEHAVIOR_HANDLE_SYNC) &&
+		    ei_event_get_type(e) == EI_EVENT_SYNC) {
+			log_debug(peck, "Skipping over sync event\n");
+			continue;
+		}
+
+		munit_errorf("Expected empty event queue, have: %s, line %d\n",
+			     peck_ei_event_name(e), lineno);
 	}
 }
 
 void
-peck_assert_no_eis_events(struct eis *eis)
+_peck_assert_no_eis_events(struct eis *eis, int lineno)
 {
 	struct peck *peck = eis_get_user_data(eis);
 
@@ -1254,14 +1275,21 @@ peck_assert_no_eis_events(struct eis *eis)
 		if (!e)
 			return;
 
-		if (peck && flag_is_set(peck->eis_behavior, PECK_EIS_BEHAVIOR_HANDLE_FRAME) &&
-		    eis_event_get_type(e) == EIS_EVENT_FRAME) {
-			log_debug(peck, "Skipping over frame event\n");
-			continue;
+		if (peck) {
+			if (flag_is_set(peck->eis_behavior, PECK_EIS_BEHAVIOR_HANDLE_FRAME) &&
+			    eis_event_get_type(e) == EIS_EVENT_FRAME) {
+				log_debug(peck, "Skipping over frame event\n");
+				continue;
+			}
+			if (flag_is_set(peck->eis_behavior, PECK_EIS_BEHAVIOR_HANDLE_SYNC) &&
+			    eis_event_get_type(e) == EIS_EVENT_SYNC) {
+				log_debug(peck, "Skipping over sync event\n");
+				continue;
+			}
 		}
 
-		munit_errorf("Expected empty event queue, have: %s\n",
-			     peck_eis_event_name(e));
+		munit_errorf("Expected empty event queue, have: %s, line %d\n",
+			     peck_eis_event_name(e), lineno);
 	}
 }
 
@@ -1279,6 +1307,14 @@ _peck_ei_next_event(struct ei *ei, enum ei_event_type type, int lineno)
 		}
 	}
 
+	if (flag_is_set(peck->ei_behavior, PECK_EI_BEHAVIOR_HANDLE_SYNC)) {
+		while (event && ei_event_get_type(event) == EI_EVENT_SYNC) {
+			ei_event_unref(event);
+			log_debug(peck, "Skipping over sync event\n");
+			event = ei_get_event(ei);
+		}
+	}
+
 	if (!event)
 		munit_errorf("Expected ei event type %s, got none, line %d\n",
 			     peck_ei_event_type_name(type),
@@ -1309,12 +1345,24 @@ _peck_eis_next_event(struct eis *eis, enum eis_event_type type, int lineno)
 		}
 	}
 
+	if (flag_is_set(peck->eis_behavior, PECK_EIS_BEHAVIOR_HANDLE_SYNC)) {
+		while (event && eis_event_get_type(event) == EIS_EVENT_SYNC) {
+			eis_event_unref(event);
+			log_debug(peck, "Skipping over sync event\n");
+			event = eis_get_event(eis);
+		}
+	}
+
 	if (!event)
 		munit_errorf("Expected EIS event type %s, got none, line %d\n",
 			     peck_eis_event_type_name(type),
 			     lineno);
-	munit_assert_string_equal(peck_eis_event_name(event),
-				  peck_eis_event_type_name(type));
+	if (eis_event_get_type(event) != type) {
+		munit_errorf("Expected EIS event type %s, got %s, line %d\n",
+			     peck_eis_event_type_name(type),
+			     peck_eis_event_type_name(eis_event_get_type(event)),
+			     lineno);
+	}
 
 	log_debug(peck, "EIS passing %s to caller, line %d\n", peck_eis_event_name(event), lineno);
 
@@ -1377,6 +1425,8 @@ peck_ei_event_type_name(enum ei_event_type type)
 		CASE_STRING(DEVICE_PAUSED);
 		CASE_STRING(DEVICE_RESUMED);
 		CASE_STRING(KEYBOARD_MODIFIERS);
+		CASE_STRING(PONG);
+		CASE_STRING(SYNC);
 		CASE_STRING(FRAME);
 		CASE_STRING(DEVICE_START_EMULATING);
 		CASE_STRING(DEVICE_STOP_EMULATING);
@@ -1413,6 +1463,8 @@ peck_eis_event_type_name(enum eis_event_type type)
 		CASE_STRING(CLIENT_DISCONNECT);
 		CASE_STRING(SEAT_BIND);
 		CASE_STRING(DEVICE_CLOSED);
+		CASE_STRING(PONG);
+		CASE_STRING(SYNC);
 		CASE_STRING(DEVICE_START_EMULATING);
 		CASE_STRING(DEVICE_STOP_EMULATING);
 		CASE_STRING(POINTER_MOTION);
