@@ -289,11 +289,13 @@ MUNIT_TEST(test_ei_device_close)
 	_unref_(peck) *peck = peck_new();
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_CLIENT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_BIND_SEAT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_DEFAULT_SEAT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_DEVICES);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_NONE);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTODEVICES);
 
 	peck_dispatch_until_stable(peck);
@@ -909,6 +911,7 @@ MUNIT_TEST(test_ei_device_pointer_abs)
 
 		/* Don't auto-handle the DEVICE_CLOSED event */
 		peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+		peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	}
 
 	with_client(peck) {
@@ -1355,6 +1358,95 @@ MUNIT_TEST(test_ei_device_touch)
 	return MUNIT_OK;
 }
 
+MUNIT_TEST(test_ei_device_touch_cancel)
+{
+	_unref_(peck) *peck = peck_new();
+	struct ei_device *device = NULL;
+
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_TOUCH);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTODEVICES);
+	peck_dispatch_until_stable(peck);
+
+	with_client(peck) {
+		device = peck_ei_get_default_touch(peck);
+		/* We know our default device has one region */
+		_unref_(ei_touch) *t = ei_device_touch_new(device);
+		ei_touch_down(t, 1, 2);
+		ei_device_frame(device, peck_ei_now(peck));
+		ei_touch_motion(t, 200, 500);
+		ei_device_frame(device, peck_ei_now(peck));
+		ei_touch_cancel(t);
+		ei_device_frame(device, peck_ei_now(peck));
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_server(peck) {
+		_unref_(eis_event) *down = peck_eis_touch_down(eis, 1, 2);
+		uint32_t tid = eis_event_touch_get_id(down);
+
+		_unref_(eis_event) *motion = peck_eis_touch_motion(eis, 200, 500);
+		munit_assert_uint32(eis_event_touch_get_id(motion), ==, tid);
+
+		_unref_(eis_event) *up = peck_eis_touch_up(eis);
+		munit_assert_uint32(eis_event_touch_get_id(up), ==, tid);
+		munit_assert_true(eis_event_touch_get_is_cancel(up));
+
+		peck_assert_no_eis_events(eis);
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_client(peck) {
+		/* up + cancel, latter is ignored */
+		_unref_(ei_touch) *t = ei_device_touch_new(device);
+		ei_touch_down(t, 100, 200);
+		ei_device_frame(device, peck_ei_now(peck));
+		ei_touch_motion(t, 300, 400);
+		ei_device_frame(device, peck_ei_now(peck));
+		ei_touch_up(t);
+		with_nonfatal_ei_bug(peck)
+			ei_touch_cancel(t); /* ignored */
+		ei_device_frame(device, peck_ei_now(peck));
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_server(peck) {
+		_unref_(eis_event) *down = peck_eis_touch_down(eis, 100, 200);
+		_unref_(eis_event) *motion = peck_eis_touch_motion(eis, 300, 400);
+		_unref_(eis_event) *up = peck_eis_touch_up(eis);
+		munit_assert_false(eis_event_touch_get_is_cancel(up));
+		peck_assert_no_eis_events(eis);
+	}
+
+	with_client(peck) {
+		/* cancel + up, latter is ignored */
+		_unref_(ei_touch) *t = ei_device_touch_new(device);
+		ei_touch_down(t, 100, 200);
+		ei_device_frame(device, peck_ei_now(peck));
+		ei_touch_motion(t, 300, 400);
+		ei_device_frame(device, peck_ei_now(peck));
+		ei_touch_cancel(t);
+		with_nonfatal_ei_bug(peck)
+			ei_touch_up(t); /* ignored */
+		ei_device_frame(device, peck_ei_now(peck));
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_server(peck) {
+		_unref_(eis_event) *down = peck_eis_touch_down(eis, 100, 200);
+		_unref_(eis_event) *motion = peck_eis_touch_motion(eis, 300, 400);
+		_unref_(eis_event) *up = peck_eis_touch_up(eis);
+		munit_assert_true(eis_event_touch_get_is_cancel(up));
+		peck_assert_no_eis_events(eis);
+	}
+
+	return MUNIT_OK;
+}
+
 MUNIT_TEST(test_ei_device_multitouch)
 {
 	_unref_(peck) *peck = peck_new();
@@ -1452,6 +1544,7 @@ MUNIT_TEST(test_ei_keymap_set)
 	_unref_(eis_keymap) *keymap = NULL;
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_CLIENT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_DEFAULT_SEAT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_BIND_SEAT);
@@ -1565,6 +1658,7 @@ MUNIT_TEST(test_ei_frame_timestamp)
 	uint64_t ts1 = 0, ts2 = 0;
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_KEYBOARD);
@@ -1613,6 +1707,7 @@ MUNIT_TEST(test_ei_no_empty_frames)
 	_unref_(peck) *peck = peck_new();
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_KEYBOARD);
@@ -1645,6 +1740,7 @@ MUNIT_TEST(test_ei_buffered_frame)
 	_unref_(peck) *peck = peck_new();
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_KEYBOARD);
@@ -1700,6 +1796,8 @@ MUNIT_TEST(test_ei_flush_frame)
 	_unref_(peck) *peck = peck_new();
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_KEYBOARD);
@@ -1766,7 +1864,9 @@ MUNIT_TEST(test_passive_ei_device_start_stop_emulating)
 {
 	_unref_(peck) *peck = peck_new_context(PECK_EI_RECEIVER);
 
-	peck_enable_eis_behavior(peck, 	PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_DEFAULT_SEAT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_CLIENT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_BIND_SEAT);
@@ -1774,6 +1874,7 @@ MUNIT_TEST(test_passive_ei_device_start_stop_emulating)
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_NONE);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_CONNECT);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTOSEAT);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_ADDED_POINTER);
@@ -1805,7 +1906,9 @@ MUNIT_TEST(test_passive_ei_device_stop_emulating_when_removing)
 {
 	_unref_(peck) *peck = peck_new_context(PECK_EI_RECEIVER);
 
-	peck_enable_eis_behavior(peck, 	PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_DEFAULT_SEAT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_CLIENT);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_BIND_SEAT);
@@ -1813,6 +1916,7 @@ MUNIT_TEST(test_passive_ei_device_stop_emulating_when_removing)
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_NONE);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_CONNECT);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTOSEAT);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_ADDED_POINTER);
@@ -2611,6 +2715,114 @@ MUNIT_TEST(test_passive_ei_device_touch)
 	return MUNIT_OK;
 }
 
+/* same as test_ei_device_touch_cancel but for a passive context */
+MUNIT_TEST(test_passive_ei_device_touch_cancel)
+{
+	_unref_(peck) *peck = peck_new_context(PECK_EI_RECEIVER);
+	struct eis_device *device = NULL;
+
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_TOUCH);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTODEVICES);
+	peck_dispatch_until_stable(peck);
+
+	uint32_t sequence = 123;
+
+	with_server(peck) {
+		device = peck_eis_get_default_touch(peck);
+		eis_device_start_emulating(device, sequence);
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_client(peck) {
+		_unref_(ei_event) *start =
+			peck_ei_next_event(ei, EI_EVENT_DEVICE_START_EMULATING);
+		munit_assert_uint(ei_event_emulating_get_sequence(start), ==, sequence);
+	}
+
+	with_server(peck) {
+		device = peck_eis_get_default_touch(peck);
+
+		_unref_(eis_touch) *t = eis_device_touch_new(device);
+		eis_touch_down(t, 1, 2);
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_touch_motion(t, 200, 500);
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_touch_cancel(t);
+		eis_device_frame(device, peck_eis_now(peck));
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_client(peck) {
+		_unref_(ei_event) *down = peck_ei_touch_down(ei, 1, 2);
+		uint32_t tid = ei_event_touch_get_id(down);
+
+		_unref_(ei_event) *motion = peck_ei_touch_motion(ei, 200, 500);
+		munit_assert_uint32(ei_event_touch_get_id(motion), ==, tid);
+
+		_unref_(ei_event) *up = peck_ei_touch_up(ei);
+		munit_assert_uint32(ei_event_touch_get_id(up), ==, tid);
+		munit_assert_true(ei_event_touch_get_is_cancel(up));
+		peck_assert_no_ei_events(ei);
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_server(peck) {
+		/* up + cancel, latter is ignored */
+		_unref_(eis_touch) *t = eis_device_touch_new(device);
+		eis_touch_down(t, 100, 200);
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_touch_motion(t, 300, 400);
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_touch_up(t);
+		eis_device_frame(device, peck_eis_now(peck));
+		with_nonfatal_eis_bug(peck)
+			eis_touch_cancel(t); /* ignored */
+		eis_device_frame(device, peck_eis_now(peck));
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_client(peck) {
+		_unref_(ei_event) *down = peck_ei_touch_down(ei, 100, 200);
+		_unref_(ei_event) *motion = peck_ei_touch_motion(ei, 300, 400);
+		_unref_(ei_event) *up = peck_ei_touch_up(ei);
+		munit_assert_false(ei_event_touch_get_is_cancel(up));
+		peck_assert_no_ei_events(ei);
+	}
+
+	with_server(peck) {
+		/* cancel + up, latter is ignored */
+		_unref_(eis_touch) *t = eis_device_touch_new(device);
+		eis_touch_down(t, 100, 200);
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_touch_motion(t, 300, 400);
+		eis_device_frame(device, peck_eis_now(peck));
+		eis_touch_cancel(t);
+		eis_device_frame(device, peck_eis_now(peck));
+		with_nonfatal_eis_bug(peck)
+			eis_touch_up(t); /* ignored */
+		eis_device_frame(device, peck_eis_now(peck));
+	}
+
+	peck_dispatch_until_stable(peck);
+
+	with_client(peck) {
+		_unref_(ei_event) *down = peck_ei_touch_down(ei, 100, 200);
+		_unref_(ei_event) *motion = peck_ei_touch_motion(ei, 300, 400);
+		_unref_(ei_event) *up = peck_ei_touch_up(ei);
+		munit_assert_true(ei_event_touch_get_is_cancel(up));
+		peck_assert_no_ei_events(ei);
+	}
+
+	return MUNIT_OK;
+}
+
 /* same as test_ei_device_multitouch but for a passive context */
 MUNIT_TEST(test_passive_ei_device_multitouch)
 {
@@ -2687,10 +2899,12 @@ MUNIT_TEST(test_passive_ei_frame_timestamp)
 	uint64_t ts1 = 0, ts2 = 0;
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_KEYBOARD);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_NONE);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_CONNECT);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTODEVICES);
 	peck_dispatch_until_stable(peck);
@@ -2745,10 +2959,12 @@ MUNIT_TEST(test_passive_ei_flush_frame)
 	_unref_(peck) *peck = peck_new_context(PECK_EI_RECEIVER);
 
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_NONE);
+	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ACCEPT_ALL);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_RESUME_DEVICE);
 	peck_enable_eis_behavior(peck, PECK_EIS_BEHAVIOR_ADD_KEYBOARD);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_NONE);
+	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_SYNC);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_HANDLE_CONNECT);
 	peck_enable_ei_behavior(peck, PECK_EI_BEHAVIOR_AUTODEVICES);
 	peck_dispatch_until_stable(peck);
