@@ -172,11 +172,16 @@ colorprint(const char *format, ...)
 		color = rgb(1, 1, 1) | rgb_bg(255, 127, 0);
 	}
 
-	cprintf(color, "EIS socket server: ");
+	const char prefix[] = "EIS socket server: ";
+	char buf[1024];
+
+	snprintf(buf, sizeof(buf), "%s", prefix);
+
 	va_list args;
 	va_start(args, format);
-	vprintf(format, args);
+	vsnprintf(buf + strlen(prefix), sizeof(buf) - strlen(prefix), format, args);
 	va_end(args);
+	cprintf(color, "%s", buf);
 }
 
 #if HAVE_MEMFD_CREATE && HAVE_LIBXKBCOMMON
@@ -436,7 +441,8 @@ eis_demo_server_printf_handle_event(struct eis_demo_server *server,
 		 * This is a demo server after all, so let's demo this. */
 		if (!eis_event_seat_has_capability(e, EIS_DEVICE_CAP_POINTER) &&
 		    !eis_event_seat_has_capability(e, EIS_DEVICE_CAP_POINTER_ABSOLUTE) &&
-		    !eis_event_seat_has_capability(e, EIS_DEVICE_CAP_KEYBOARD))
+		    !eis_event_seat_has_capability(e, EIS_DEVICE_CAP_KEYBOARD) &&
+		    !eis_event_seat_has_capability(e, EIS_DEVICE_CAP_TOUCH))
 			eis_seat_remove(seat);
 
 		break;
@@ -540,7 +546,24 @@ eis_demo_server_printf_handle_event(struct eis_demo_server *server,
 			   eis_event_get_time(e));
 		}
 		break;
+	case EIS_EVENT_PONG:
+		{
+		colorprint("pong\n");
+		}
+		break;
+	case EIS_EVENT_SYNC:
+		{
+		colorprint("sync\n");
+		}
+		break;
 	default:
+		/* This is a demo server and we abort to make it easy to debug
+		 * a missing implementation of some new event.
+		 *
+		 * Do not do this in a real EIS implementation, unknown
+		 * events must be passed to eis_event_unref() and otherwise
+		 * ignored.
+		 */
 		abort();
 	}
 	return 0;
